import pytest

from curl_cffi import CurlHttpVersion, CurlOpt, requests

#######################################################################################
# testing httpbin
#######################################################################################


def test_gzip():
    r = requests.get("https://httpbin.org/gzip")
    assert r.status_code == 200
    assert r.json()["gzipped"]


def test_brotli():
    r = requests.get("https://httpbin.org/brotli")
    assert r.status_code == 200
    assert r.json()["brotli"]


def test_redirect_n():
    r = requests.get("https://httpbin.org/absolute-redirect/3")
    assert r.status_code == 200


def test_relative_redirect_n():
    r = requests.get("https://httpbin.org/relative-redirect/3")
    assert r.status_code == 200


def test_redirect_max():
    with pytest.raises(requests.RequestsError, match="Failed"):
        requests.get("https://httpbin.org/redirect/5", max_redirects=3)


def test_imperonsate_default_headers():
    r = requests.get("http://postman-echo.com/headers", impersonate="chrome110")
    headers = r.json()
    assert "Mozilla" in headers["headers"]["user-agent"]
    r = requests.get(
        "http://httpbin.org/headers",
        impersonate="chrome110",
        default_headers=False,
    )
    headers = r.json()
    assert "user-agent" not in headers["headers"]


def test_curl_options():
    url = "https://postman-echo.com/headers"
    r = requests.get(url, impersonate="chrome110", default_headers=False)
    headers = r.json()
    print(headers)
    assert r.http_version == CurlHttpVersion.V2_0
    r = requests.get(
        url,
        curl_options={CurlOpt.HTTP_VERSION: CurlHttpVersion.V1_1},
        impersonate="chrome110",
        default_headers=False,
    )
    headers = r.json()
    assert r.http_version == CurlHttpVersion.V1_1
    print(headers)


def test_http_version():
    url = "https://postman-echo.com/headers"
    r = requests.get(url, impersonate="chrome110", http_version=CurlHttpVersion.V1_1)
    headers = r.json()
    assert r.http_version == CurlHttpVersion.V1_1
    print(headers)



import json
import os
import platform
import shutil
import struct
import tempfile
from pathlib import Path
from urllib.request import urlretrieve

from cffi import FFI

# this is the upstream libcurl-impersonate version
__version__ = "0.7.0"

tmpdir = None


def detect_arch():
    with open(Path(__file__).parent.parent / "libs.json") as f:
        archs = json.loads(f.read())

    uname = platform.uname()
    glibc_flavor = "gnueabihf" if uname.machine in ["armv7l", "armv6l"] else "gnu"

    libc, _ = platform.libc_ver()
    # https://github.com/python/cpython/issues/87414
    libc = glibc_flavor if libc == "glibc" else "musl"
    pointer_size = struct.calcsize("P") * 8

    for arch in archs:
        if (
            arch["system"] == uname.system
            and arch["machine"] == uname.machine
            and arch["pointer_size"] == pointer_size
            and ("libc" not in arch or arch.get("libc") == libc)
        ):
            if arch["libdir"]:
                arch["libdir"] = os.path.expanduser(arch["libdir"])
            else:
                global tmpdir
                tmpdir = tempfile.TemporaryDirectory()
                arch["libdir"] = tmpdir.name
            return arch
    raise Exception(f"Unsupported arch: {uname}")


arch = detect_arch()
print(f"Using {arch['libdir']} to store libcurl-impersonate")


def download_libcurl():
    if (Path(arch["libdir"]) / arch["so_name"]).exists():
        print(".so files already downloaded.")
        return

    file = "libcurl-impersonate.tar.gz"
    sysname = "linux-" + arch["libc"] if arch["system"] == "Linux" else arch["sysname"]

    url = (
        f"https://github.com/yifeikong/curl-impersonate/releases/download/"
        f"v{__version__}/libcurl-impersonate-v{__version__}"
        f".{arch['so_arch']}-{sysname}.tar.gz"
    )

    print(f"Downloading libcurl-impersonate-chrome from {url}...")
    urlretrieve(url, file)

    print("Unpacking downloaded files...")
    os.makedirs(arch["libdir"], exist_ok=True)
    shutil.unpack_archive(file, arch["libdir"])


def get_curl_archives():
    if arch["system"] == "Linux" and arch.get("link_type") == "static":
        # note that the order of libraries matters
        # https://stackoverflow.com/a/36581865
        return [
            f"{arch['libdir']}/libcurl-impersonate-chrome.a",
            f"{arch['libdir']}/libssl.a",
            f"{arch['libdir']}/libcrypto.a",
            f"{arch['libdir']}/libz.a",
            f"{arch['libdir']}/libzstd.a",
            f"{arch['libdir']}/libnghttp2.a",
            f"{arch['libdir']}/libbrotlidec.a",
            f"{arch['libdir']}/libbrotlienc.a",
            f"{arch['libdir']}/libbrotlicommon.a",
        ]
    else:
        return []


def get_curl_libraries():
    if arch["system"] == "Windows":
        return ["libcurl"]
    elif arch["system"] == "Darwin" or (
        arch["system"] == "Linux" and arch.get("link_type") == "dynamic"
    ):
        return ["curl-impersonate-chrome"]
    else:
        return []


ffibuilder = FFI()
system = platform.system()
root_dir = Path(__file__).parent.parent
download_libcurl()


ffibuilder.set_source(
    "curl_cffi._wrapper",
    """
        #include "shim.h"
    """,
    # FIXME from `curl-impersonate`
    libraries=get_curl_libraries(),
    extra_objects=get_curl_archives(),
    library_dirs=[arch["libdir"]],
    source_extension=".c",
    include_dirs=[
        str(root_dir / "include"),
        str(root_dir / "ffi"),
    ],
    sources=[
        str(root_dir / "ffi/shim.c"),
    ],
    extra_compile_args=(["-Wno-implicit-function-declaration"] if system == "Darwin" else []),
)

with open(root_dir / "ffi/cdef.c") as f:
    cdef_content = f.read()
    ffibuilder.cdef(cdef_content)


if __name__ == "__main__":
    ffibuilder.compile(verbose=False)



import pytest

from curl_cffi import requests

JA3_URL = "https://tls.browserleaks.com/json"
# Copied from my browser on macOS
CHROME_JA3_HASH = "53ff64ddf993ca882b70e1c82af5da49"
# Edge 101 is the same as Chrome 101
EDGE_JA3_HASH = "53ff64ddf993ca882b70e1c82af5da49"
# Same as safari 16.x
SAFARI_JA3_HASH = "8468a1ef6cb71b13e1eef8eadf786f7d"


def test_not_impersonate():
    r = requests.get(JA3_URL)
    assert r.json()["ja3_hash"] != CHROME_JA3_HASH


def test_impersonate():
    r = requests.get(JA3_URL, impersonate="chrome101")
    assert r.json()["ja3_hash"] == CHROME_JA3_HASH


def test_impersonate_edge():
    r = requests.get(JA3_URL, impersonate=requests.BrowserType.edge101)
    assert r.json()["ja3_hash"] == EDGE_JA3_HASH


def test_impersonate_safari():
    r = requests.get(JA3_URL, impersonate="safari15_5")
    assert r.json()["ja3_hash"] == SAFARI_JA3_HASH


def test_impersonate_unknown():
    with pytest.raises(requests.RequestsError, match="not supported"):
        requests.get(JA3_URL, impersonate="unknown")



import pytest

from curl_cffi.requests.cookies import Cookies, CurlMorsel
from curl_cffi.requests.errors import CookieConflict, RequestsError


def test_cookies_conflict():
    c = Cookies()
    c.set("foo", "bar", domain="example.com")
    c.set("foo", "baz", domain="test.local")
    with pytest.raises(CookieConflict):
        c.get("foo")


def test_cookies_conflict_on_subdomain():
    c = Cookies()
    c.set("foo", "bar", domain=".example.com")
    c.set("foo", "baz", domain="a.example.com")
    assert c.get("foo") in ("bar", "baz")


def test_cookies_conflict_but_same():
    c = Cookies()
    c = Cookies()
    c.set("foo", "bar", domain="example.com")
    c.set("foo", "bar", domain="test.local")
    assert c.get("foo") == "bar"


def test_curl_format_with_hostname():
    m = CurlMorsel(name="foo", value="bar", hostname="example.com")
    assert m.to_curl_format() == "example.com\tFALSE\t/\tFALSE\t0\tfoo\tbar"
    m = CurlMorsel(name="foo", value="bar", hostname="example.com", secure=True)
    assert m.to_curl_format() == "example.com\tFALSE\t/\tTRUE\t0\tfoo\tbar"
    m = CurlMorsel(name="foo", value="bar", hostname="example.com", path="/path")
    assert m.to_curl_format() == "example.com\tFALSE\t/path\tFALSE\t0\tfoo\tbar"


def test_curl_format_without_hostname():
    m = CurlMorsel(name="foo", value="bar")
    with pytest.raises(RequestsError):
        m.to_curl_format()



import base64
import json
from io import BytesIO
from typing import cast

import pytest

from curl_cffi import Curl, CurlError, CurlInfo, CurlOpt

#######################################################################################
# testing setopt
#######################################################################################


def test_get(server):
    c = Curl()
    c.setopt(CurlOpt.URL, str(server.url).encode())
    c.perform()


def test_post(server):
    c = Curl()
    url = str(server.url.copy_with(path="/echo_body"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.POST, 1)
    c.setopt(CurlOpt.POSTFIELDS, b"foo=bar")
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    assert buffer.getvalue() == b"foo=bar"


def test_put(server):
    c = Curl()
    c.setopt(CurlOpt.URL, str(server.url).encode())
    c.setopt(CurlOpt.CUSTOMREQUEST, b"PUT")
    c.perform()


def test_delete(server):
    c = Curl()
    c.setopt(CurlOpt.URL, str(server.url).encode())
    c.setopt(CurlOpt.CUSTOMREQUEST, b"DELETE")
    c.perform()


def test_post_data_with_size(server):
    c = Curl()
    url = str(server.url.copy_with(path="/echo_body"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.CUSTOMREQUEST, b"POST")
    c.setopt(CurlOpt.POSTFIELDS, b"\0" * 7)
    c.setopt(CurlOpt.POSTFIELDSIZE, 7)
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    assert buffer.getvalue() == b"\0" * 7


def test_headers(server):
    c = Curl()
    url = str(server.url.copy_with(path="/echo_headers"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.HTTPHEADER, [b"Foo: bar"])
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    headers = json.loads(buffer.getvalue().decode())
    assert headers["Foo"][0] == "bar"

    # https://github.com/yifeikong/curl_cffi/issues/16
    c.setopt(CurlOpt.HTTPHEADER, [b"Foo: baz"])
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    headers = json.loads(buffer.getvalue().decode())
    assert headers["Foo"][0] == "baz"


def test_proxy_headers(server):
    # XXX only tests that proxy header is not present for target server, should add
    # tests that verifies proxy headers are sent to proxy server.
    c = Curl()
    url = str(server.url.copy_with(path="/echo_headers"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.PROXYHEADER, [b"Foo: bar"])
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    headers = json.loads(buffer.getvalue().decode())
    assert "Foo" not in headers

    # https://github.com/yifeikong/curl_cffi/issues/16
    c.setopt(CurlOpt.PROXYHEADER, [b"Foo: baz"])
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    headers = json.loads(buffer.getvalue().decode())
    assert "Foo" not in headers


def test_write_function_memory_leak(server):
    c = Curl()
    for _ in range(10):
        url = str(server.url.copy_with(path="/echo_headers"))
        c.setopt(CurlOpt.URL, url.encode())
        c.setopt(CurlOpt.HTTPHEADER, [b"Foo: bar"])
        buffer = BytesIO()
        c.setopt(CurlOpt.WRITEDATA, buffer)
        c.perform()
    assert c._write_handle is None


def test_write_function(server):
    c = Curl()
    url = str(server.url.copy_with(path="/echo_body"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.POST, 1)
    c.setopt(CurlOpt.POSTFIELDS, b"foo=bar")

    buffer = BytesIO()

    def write(data: bytes):
        buffer.write(data)
        return len(data)

    c.setopt(CurlOpt.WRITEFUNCTION, write)
    c.perform()
    assert buffer.getvalue() == b"foo=bar"


def test_cookies(server):
    c = Curl()
    url = str(server.url.copy_with(path="/echo_cookies"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.COOKIE, b"foo=bar")
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    cookies = json.loads(buffer.getvalue().decode())
    # print(cookies)
    assert cookies["foo"] == "bar"


def test_auth(server):
    c = Curl()
    url = str(server.url.copy_with(path="/echo_headers"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.USERNAME, b"foo")
    c.setopt(CurlOpt.PASSWORD, b"bar")
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    headers = json.loads(buffer.getvalue().decode())
    assert headers["Authorization"][0] == f"Basic {base64.b64encode(b'foo:bar').decode()}"


def test_timeout(server):
    c = Curl()
    url = str(server.url.copy_with(path="/slow_response"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.TIMEOUT_MS, 100)
    with pytest.raises(CurlError, match=r"curl: \(28\)"):
        c.perform()


def test_repeated_headers_after_error(server):
    c = Curl()
    url = str(server.url.copy_with(path="/slow_response"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.TIMEOUT_MS, 100)
    c.setopt(CurlOpt.HTTPHEADER, [b"Foo: bar"])
    with pytest.raises(CurlError, match=r"curl: \(28\)"):
        c.perform()

    # another request
    url = str(server.url.copy_with(path="/echo_headers"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.HTTPHEADER, [b"Foo: bar"])
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    headers = json.loads(buffer.getvalue().decode())
    assert len(headers["Foo"]) == 1
    # print(headers)


def test_follow_redirect(server):
    c = Curl()
    url = str(server.url.copy_with(path="/redirect_301"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.FOLLOWLOCATION, 1)
    c.perform()
    assert c.getinfo(CurlInfo.RESPONSE_CODE) == 200


def test_not_follow_redirect(server):
    c = Curl()
    url = str(server.url.copy_with(path="/redirect_301"))
    c.setopt(CurlOpt.URL, url.encode())
    c.perform()
    assert c.getinfo(CurlInfo.RESPONSE_CODE) == 301


def test_http_proxy_changed_path(server):
    c = Curl()
    proxy_url = str(server.url).rstrip("/")
    print("proxy url", proxy_url)
    c.setopt(CurlOpt.URL, b"http://example.org")
    c.setopt(CurlOpt.PROXY, proxy_url.encode())
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    rsp = json.loads(buffer.getvalue().decode())
    assert rsp["Hello"] == "http_proxy!"


def test_https_proxy_using_connect(server):
    c = Curl()
    proxy_url = str(server.url)
    c.setopt(CurlOpt.URL, b"https://example.org")
    c.setopt(CurlOpt.PROXY, proxy_url.encode())
    c.setopt(CurlOpt.HTTPPROXYTUNNEL, 1)
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    with pytest.raises(CurlError, match=r"curl: \(35\)"):
        c.perform()


def test_verify(https_server):
    c = Curl()
    url = str(https_server.url)
    c.setopt(CurlOpt.URL, url.encode())
    with pytest.raises(CurlError, match="SSL certificate problem"):
        c.perform()


def test_verify_false(https_server):
    c = Curl()
    url = str(https_server.url)
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.SSL_VERIFYPEER, 0)
    c.setopt(CurlOpt.SSL_VERIFYHOST, 0)
    c.perform()


def test_referer(server):
    c = Curl()
    url = str(server.url.copy_with(path="/echo_headers"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.REFERER, b"http://example.org")
    buffer = BytesIO()
    c.setopt(CurlOpt.WRITEDATA, buffer)
    c.perform()
    headers = json.loads(buffer.getvalue().decode())
    assert headers["Referer"][0] == "http://example.org"


#######################################################################################
# testing getinfo
#######################################################################################


def test_effective_url(server):
    c = Curl()
    url = str(server.url.copy_with(path="/redirect_301"))
    c.setopt(CurlOpt.URL, url.encode())
    c.setopt(CurlOpt.FOLLOWLOCATION, 1)
    c.perform()
    assert c.getinfo(CurlInfo.EFFECTIVE_URL) == str(server.url).encode()


def test_status_code(server):
    c = Curl()
    url = str(server.url)
    c.setopt(CurlOpt.URL, url.encode())
    c.perform()
    assert c.getinfo(CurlInfo.RESPONSE_CODE) == 200


def test_response_headers(server):
    c = Curl()
    url = str(server.url.copy_with(path="/set_headers"))
    c.setopt(CurlOpt.URL, url.encode())
    buffer = BytesIO()
    c.setopt(CurlOpt.HEADERDATA, buffer)
    c.perform()
    headers = buffer.getvalue().decode()
    for line in headers.splitlines():
        if line.startswith("x-test"):
            assert line.startswith("x-test: test")


def test_response_cookies(server):
    c = Curl()
    url = str(server.url.copy_with(path="/set_cookies"))
    c.setopt(CurlOpt.URL, url.encode())
    buffer = BytesIO()
    c.setopt(CurlOpt.HEADERDATA, buffer)
    c.perform()
    headers = buffer.getvalue()
    cookie = c.parse_cookie_headers(headers.splitlines())
    for name, morsel in cookie.items():
        if name == "foo":
            assert morsel.value == "bar"


def test_elapsed(server):
    c = Curl()
    url = str(server.url)
    c.setopt(CurlOpt.URL, url.encode())
    c.perform()
    assert cast(int, c.getinfo(CurlInfo.TOTAL_TIME)) > 0


def test_reason(server):
    c = Curl()
    url = str(server.url)
    c.setopt(CurlOpt.URL, url.encode())
    buffer = BytesIO()
    c.setopt(CurlOpt.HEADERDATA, buffer)
    c.perform()
    headers = buffer.getvalue()
    headers = headers.splitlines()
    assert c.get_reason_phrase(headers[0]) == b"OK"


def test_resolve(server):
    c = Curl()
    url = "http://example.com:8000"
    c.setopt(CurlOpt.RESOLVE, ["example.com:8000:127.0.0.1"])
    c.setopt(CurlOpt.URL, url)
    c.perform()


def test_duphandle(server):
    c = Curl()
    c.setopt(CurlOpt.URL, str(server.url.copy_with(path="/redirect_loop")).encode())
    c.setopt(CurlOpt.FOLLOWLOCATION, 1)
    c.setopt(CurlOpt.MAXREDIRS, 2)
    c = c.duphandle()
    with pytest.raises(CurlError):
        c.perform()



__all__ = [
    "Curl",
    "AsyncCurl",
    "CurlMime",
    "CurlError",
    "CurlInfo",
    "CurlOpt",
    "CurlMOpt",
    "CurlECode",
    "CurlHttpVersion",
    "CurlSslVersion",
    "CurlWsFlag",
    "ffi",
    "lib",
]

import _cffi_backend  # noqa: F401  # required by _wrapper

from .__version__ import __curl_version__, __description__, __title__, __version__  # noqa: F401

# This line includes _wrapper.so into the wheel
from ._wrapper import ffi, lib
from .aio import AsyncCurl
from .const import CurlECode, CurlHttpVersion, CurlInfo, CurlMOpt, CurlOpt, CurlWsFlag, CurlSslVersion
from .curl import Curl, CurlError, CurlMime



import asyncio
import base64
import json
from contextlib import suppress

import pytest

from curl_cffi.requests import AsyncSession, RequestsError
from curl_cffi.requests.errors import SessionClosed


async def test_get(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url))
        assert r.status_code == 200


def test_create_session_out_of_async(server):
    s = AsyncSession()

    async def get():
        r = await s.get(str(server.url))
        assert r.status_code == 200

    asyncio.run(get())


async def test_post_dict(server):
    async with AsyncSession() as s:
        r = await s.post(str(server.url.copy_with(path="/echo_body")), data={"foo": "bar"})
        assert r.status_code == 200
        assert r.content == b"foo=bar"


async def test_post_str(server):
    async with AsyncSession() as s:
        r = await s.post(str(server.url.copy_with(path="/echo_body")), data='{"foo": "bar"}')
        assert r.status_code == 200
        assert r.content == b'{"foo": "bar"}'


async def test_post_json(server):
    async with AsyncSession() as s:
        r = await s.post(str(server.url.copy_with(path="/echo_body")), json={"foo": "bar"})
        assert r.status_code == 200
        assert r.content == b'{"foo":"bar"}'


async def test_put_json(server):
    async with AsyncSession() as s:
        r = await s.put(str(server.url.copy_with(path="/echo_body")), json={"foo": "bar"})
        assert r.status_code == 200
        assert r.content == b'{"foo":"bar"}'


async def test_delete(server):
    async with AsyncSession() as s:
        r = await s.delete(str(server.url.copy_with(path="/echo_body")))
        assert r.status_code == 200


async def test_options(server):
    async with AsyncSession() as s:
        r = await s.options(str(server.url.copy_with(path="/echo_body")))
        assert r.status_code == 200


async def test_base_url(server):
    async with AsyncSession(
        base_url=str(server.url.copy_with(path="/a/b", params={"foo": "bar"}))
    ) as s:
        # target path is empty
        r = await s.get("")
        assert r.url == s.base_url

        # target path only has params
        r = await s.get("", params={"hello": "world"})
        assert r.url == str(server.url.copy_with(path="/a/b", params={"hello": "world"}))

        # target path is a relative path without starting /
        r = await s.get("x")
        assert r.url == str(server.url.copy_with(path="/a/x"))
        r = await s.get("x", params={"hello": "world"})
        assert r.url == str(server.url.copy_with(path="/a/x", params={"hello": "world"}))

        # target path is a relative path with starting /
        r = await s.get("/x")
        assert r.url == str(server.url.copy_with(path="/x"))
        r = await s.get("/x", params={"hello": "world"})
        assert r.url == str(server.url.copy_with(path="/x", params={"hello": "world"}))

        # target path is an absolute url
        r = await s.get(str(server.url.copy_with(path="/x/y")))
        assert r.url == str(server.url.copy_with(path="/x/y"))


async def test_params(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/echo_params")), params={"foo": "bar"})
        assert r.status_code == 200
        assert r.content == b'{"params": {"foo": ["bar"]}}'


async def test_update_params(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/echo_params?foo=z")), params={"foo": "bar"})
        assert r.status_code == 200
        assert r.content == b'{"params": {"foo": ["bar"]}}'


async def test_headers(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/echo_headers")), headers={"foo": "bar"})
        headers = r.json()
        assert headers["Foo"][0] == "bar"


async def test_cookies(server):
    async with AsyncSession() as s:
        r = await s.get(
            str(server.url.copy_with(path="/echo_cookies")),
            cookies={"foo": "bar", "hello": "world"},
        )
        cookies = r.json()
        assert cookies["foo"] == "bar"


async def test_auth(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/echo_headers")), auth=("foo", "bar"))
        assert r.status_code == 200
        assert r.json()["Authorization"][0] == f"Basic {base64.b64encode(b'foo:bar').decode()}"


async def test_timeout(server):
    with pytest.raises(RequestsError):
        async with AsyncSession() as s:
            await s.get(str(server.url.copy_with(path="/slow_response")), timeout=0.1)


async def test_not_follow_redirects(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/redirect_301")), allow_redirects=False)
        assert r.status_code == 301
        assert r.redirect_count == 0
        assert r.content == b"Redirecting..."


async def test_follow_redirects(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/redirect_301")), allow_redirects=True)
        assert r.status_code == 200
        assert r.redirect_count == 1


async def test_verify(https_server):
    async with AsyncSession() as s:
        with pytest.raises(RequestsError, match="SSL certificate problem"):
            await s.get(str(https_server.url), verify=True)


async def test_verify_false(https_server):
    async with AsyncSession() as s:
        r = await s.get(str(https_server.url), verify=False)
        assert r.status_code == 200


async def test_referer(server):
    async with AsyncSession() as s:
        r = await s.get(
            str(server.url.copy_with(path="/echo_headers")),
            referer="http://example.com",
        )
        headers = r.json()
        assert headers["Referer"][0] == "http://example.com"


#######################################################################################
# testing response
#######################################################################################


async def test_redirect_url(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/redirect_301")), allow_redirects=True)
        assert r.url == str(server.url.copy_with(path="/"))


async def test_response_headers(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/set_headers")))
        assert r.headers.get_list("x-test") == ["test", "test2"]


async def test_response_cookies(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/set_cookies")))
        print(r.cookies)
        assert r.cookies["foo"] == "bar"


async def test_elapsed(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/slow_response")))
        assert r.elapsed > 0.1


async def test_reason(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/redirect_301")), allow_redirects=False)
        assert r.reason == "Moved Permanently"
        r = await s.get(str(server.url.copy_with(path="/redirect_301")), allow_redirects=True)
        assert r.status_code == 200
        assert r.reason == "OK"


#######################################################################################
# testing session
#######################################################################################


async def test_session_update_parms(server):
    async with AsyncSession(params={"old": "day"}) as s:
        r = await s.get(str(server.url.copy_with(path="/echo_params")), params={"foo": "bar"})
        assert r.content == b'{"params": {"old": ["day"], "foo": ["bar"]}}'


async def test_session_preset_cookies(server):
    async with AsyncSession(cookies={"foo": "bar"}) as s:
        # send requests with other cookies
        r = await s.get(str(server.url.copy_with(path="/echo_cookies")), cookies={"hello": "world"})
        cookies = r.json()
        # old cookies should be persisted
        assert cookies["foo"] == "bar"
        # new cookies should be added
        assert cookies["hello"] == "world"


async def test_session_cookies(server):
    async with AsyncSession() as s:
        # let the server set cookies
        r = await s.get(str(server.url.copy_with(path="/set_cookies")))
        assert s.cookies["foo"] == "bar"
        # send requests with other cookies
        r = await s.get(str(server.url.copy_with(path="/echo_cookies")), cookies={"hello": "world"})
        cookies = r.json()
        # old cookies should be persisted
        assert cookies["foo"] == "bar"
        # new cookies should be added
        assert cookies["hello"] == "world"


# https://github.com/yifeikong/curl_cffi/issues/16
async def test_session_with_headers(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url), headers={"Foo": "bar"})
        r = await s.get(str(server.url), headers={"Foo": "baz"})
        assert r.status_code == 200


async def test_session_too_many_headers(server):
    async with AsyncSession() as s:
        r = await s.get(str(server.url.copy_with(path="/echo_headers")), headers={"Foo": "1"})
        r = await s.get(str(server.url.copy_with(path="/echo_headers")), headers={"Foo": "2"})
        headers = r.json()
        assert len(headers["Foo"]) == 1
        assert headers["Foo"][0] == "2"


# https://github.com/yifeikong/curl_cffi/issues/222
async def test_closed_session_throws_error():
    async with AsyncSession() as s:
        pass

    with pytest.raises(SessionClosed):
        await s.get("https://example.com")

    with pytest.raises(SessionClosed):
        await s.post("https://example.com")

    with pytest.raises(SessionClosed):
        await s.put("https://example.com")

    with pytest.raises(SessionClosed):
        await s.delete("https://example.com")

    with pytest.raises(SessionClosed):
        await s.options("https://example.com")

    with pytest.raises(SessionClosed):
        await s.head("https://example.com")

    with pytest.raises(SessionClosed):
        await s.patch("https://example.com")

    with pytest.raises(SessionClosed):
        await s.ws_connect("wss://example.com")


# https://github.com/yifeikong/curl_cffi/issues/39
async def test_post_body_cleaned(server):
    async with AsyncSession() as s:
        # POST with body
        r = await s.post(str(server.url), json={"foo": "bar"})
        # GET request with echo_body
        r = await s.get(str(server.url.copy_with(path="/echo_body")))
        # ensure body is empty
        assert r.content == b""


async def test_timers_leak(server):
    async with AsyncSession() as sess:
        for _ in range(3):
            with suppress(Exception):
                await sess.get(str(server.url.copy_with(path="/slow_response")), timeout=0.1)
        await asyncio.sleep(0.2)
        assert len(sess.acurl._timers) == 0


#######################################################################################
# async parallel
#######################################################################################


async def test_parallel(server):
    async with AsyncSession() as s:
        rs = [
            s.get(str(server.url.copy_with(path="/echo_headers")), headers={"Foo": f"{i}"})
            for i in range(6)
        ]
        tasks = [asyncio.create_task(r) for r in rs]
        rs = await asyncio.gather(*tasks)
        for idx, r in enumerate(rs):
            assert r.status_code == 200
            assert r.json()["Foo"][0] == str(idx)


async def test_stream_iter_content(server):
    async with AsyncSession() as s:
        url = str(server.url.copy_with(path="/stream"))
        async with s.stream("GET", url, params={"n": "20"}) as r:
            async for chunk in r.aiter_content():
                assert b"path" in chunk


async def test_stream_iter_content_break(server):
    async with AsyncSession() as s:
        url = str(server.url.copy_with(path="/stream"))
        async with s.stream("GET", url, params={"n": "20"}) as r:
            idx = 0
            async for chunk in r.aiter_content():
                idx += 1
                assert b"path" in chunk
                if idx == 3:
                    break
            assert r.status_code == 200


async def test_stream_iter_lines(server):
    async with AsyncSession() as s:
        url = str(server.url.copy_with(path="/stream"))
        async with s.stream("GET", url, params={"n": "20"}) as r:
            async for chunk in r.aiter_lines():
                data = json.loads(chunk)
                assert data["path"] == "/stream"


async def test_stream_status_code(server):
    async with AsyncSession() as s:
        url = str(server.url.copy_with(path="/stream"))
        async with s.stream("GET", url, params={"n": "20"}) as r:
            assert r.status_code == 200


async def test_stream_empty_body(server):
    async with AsyncSession() as s:
        url = str(server.url.copy_with(path="/empty_body"))
        async with s.stream("GET", url) as r:
            assert r.status_code == 200


async def test_stream_atext(server):
    async with AsyncSession() as s:
        url = str(server.url.copy_with(path="/stream"))
        async with s.stream("GET", url, params={"n": "20"}) as r:
            text = await r.atext()
            chunks = text.split("\n")
            assert len(chunks) == 20



from .. import CurlError


class RequestsError(CurlError):
    """Base exception for curl_cffi.requests package"""

    def __init__(self, msg, code=0, response=None, *args, **kwargs):
        super().__init__(msg, code, *args, **kwargs)
        self.response = response


class CookieConflict(RequestsError):
    pass


class SessionClosed(RequestsError):
    pass
